/*
Developer: Dionysis "dionyziz" Zindros <dionyziz@kamibu.com>

Copyright (c) 2010, Kamibu
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:
    * Redistributions of source code must retain the above copyright
      notice, this list of conditions and the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright
      notice, this list of conditions and the following disclaimer in the
      documentation and/or other materials provided with the distribution.
    * Neither the name of Kamibu nor the
      names of its contributors may be used to endorse or promote products
      derived from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY KAMIBU ``AS IS'' AND ANY
EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL KAMIBU BE LIABLE FOR ANY
DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

var Alive = {
    CULLING: true,
    CLIP_F: 0.2,
    CLIP_B: 10000,
    LIGHT_AMBIENT: [ 50, 50, 50 ],
    frames: 0,
    createSurface: function ( canvas ) {
        var canvasSurface = canvas.getContext( '2d' );
        canvasSurface.globalCompositeOperation = 'source-over';
        return {
            canvasSurface: canvasSurface,
            width: canvas.width,
            height: canvas.height,
            clear: function () {
                this.canvasSurface.clearRect( 0, 0, this.width, this.height );
            },
            render: function ( universe ) {
                Alive.drawUniverse( universe, this );
            }
        };
    },
    createObjectCopy: function ( data ) {
        var dataCopy = [
            Alive.Math.Vector.copy( data[ 0 ] ),
            Alive.Math.Vector.copy( data[ 1 ] )
        ];
        return Alive.createObject( dataCopy );
    },
    createObject: function ( data ) {
        return {
            data: data,
            material: Alive.Material.create(),
            rotate: function ( x, y, z ) {
                Alive.Transform.rotate( this.data, [ x, y, z ] );
            },
            translate: function ( x, y, z ) {
                Alive.Transform.translate( this.data, [ x, y, z ] );
            },
            scale: function ( x, y, z ) {
                if ( typeof y == 'undefined' ) {
                    y = z = x;
                }
                Alive.Transform.scale( this.data, [ x, y, z ] );
            },
            transform: function ( matrix, includeNormals ) {
                var length;
                
                if ( includeNormals ) {
                    length = this.data[ 0 ].length;
                }
                else {
                    length = this.data[ 2 ];
                }
                
                Alive.Math.mul( this.data[ 0 ], matrix, length );
            },
            populate: function () {
                // calculate and store normals for each triangle
                // append normal points to regular points and store first
                // normal point offset
                
                var points = this.data[ 0 ];
                var triangles = this.data[ 1 ];
                var triangle;
                
                j = points.length;
                this.data[ 2 ] = j;
                for ( var i = 0; i < triangles.length; ++i ) {
                    triangle = triangles[ i ];
                    // add normal point to pointset
                    this.data[ 0 ][ j ] = Alive.Math.Vector.normal( points[ triangle[ 0 ] ], points[ triangle[ 1 ] ], points[ triangle[ 2 ] ] );
                    // associate with triangle
                    this.data[ 1 ][ i ][ 3 ] = j;
                    ++j;
                }
            },
            assignMaterial: function ( material ) {
                this.material = material;
            }
        };
    },
    Universe: {
        create: function () {
            return {
                objects: [],
                numobjects: 0,
                faceref: [],
                numfaces: 0,
                addObject: function ( obj ) {
                    this.objects[ this.numobjects ] = obj;
                    for ( var i = 0; i < obj.data[ 1 ].length; ++i ) {
                        this.faceref[ this.numfaces ] = [ this.numobjects, i ];
                        ++this.numfaces;
                    }
                    ++this.numobjects;
                }
            };
        }
    },
    Math: {
        Vector: {
            crossproduct: function( u, v ) {
                // http://en.wikipedia.org/wiki/Cross_product
                return [ u[ 1 ] * v[ 2 ] - u[ 2 ] * v[ 1 ],
                         u[ 2 ] * v[ 0 ] - u[ 0 ] * v[ 2 ],
                         u[ 0 ] * v[ 1 ] - u[ 1 ] * v[ 0 ] ];
            },
            dotproduct: function( u, v ) {
                // http://en.wikipedia.org/wiki/Dot_product
                return u[ 0 ] * v[ 0 ] + u[ 1 ] * v[ 1 ] + u[ 2 ] * v[ 2 ];
            },
            invlength: function( u ) {
                // Calculate the inverse length of vector u
                // i.e. the inverse of the distance of the vector point from the origin
                // if the vector point is at (x, y, z)
                // its distance from the origin is sqrt( x^2 + y^2 + z^2 )
                // and the inverse distance is 1 / sqrt( x^2 + y^2 + z^2 )
                // As we need to call this relatively often, while accuracy isn't
                // very important, speed is crucial. Therefore, this is an approximation
                // function which runs faster but gives less accurate results.
                
                var a = 1 / ( u[ 0 ] * u[ 0 ] + u[ 1 ] * u[ 1 ] + u[ 2 ] * u[ 2 ] );
                // var x = Math.sqrt( a ); // exact method; too slow for our needs

                // approximate using Rhapson-Newton method with 10 iterations
                // http://en.wikipedia.org/wiki/Newton's_method
                var x = a / 2; // first approximation

                for ( var i = 0; i < 10; ++i ) {
                    x = x - ( x * x - a ) / ( 2 * x );
                }

                return x;
            },
            normalize: function ( u ) {
                var L = Alive.Math.Vector.invlength( u );
                
                return [ L * u[ 0 ], L * u[ 1 ], L * u[ 2 ], 0 ];
            },
            normal: function( a, b, c ) {
                // find the normal vector of a surface on which points a, b, c lie
                
                var u = [ b[ 0 ] - a[ 0 ], b[ 1 ] - a[ 1 ], b[ 2 ] - a[ 2 ] ];
                var v = [ c[ 0 ] - a[ 0 ], c[ 1 ] - a[ 1 ], c[ 2 ] - a[ 2 ] ];
                
                // n = u x v
                return Alive.Math.Vector.normalize( Alive.Math.Vector.crossproduct( u, v ) );
            },
            copy: function( v ) {
                var b = [];
                
                for ( var i = 0; i < v.length; ++i ) {
                    b[ i ] = [
                        v[ i ][ 0 ],
                        v[ i ][ 1 ],
                        v[ i ][ 2 ],
                        v[ i ][ 3 ]
                    ];
                }
                return b;
            }
        },
        mul: function( a1, a2, n ) {
            // multiply two matrices; store the result in a1
            // http://en.wikipedia.org/wiki/Matrix_multiplication
            
            // notice that ( a2^T * a1^T )^T = a2 * a1
            // transposing is convenient to be able to store points in [ x, y, z ] form

            /*
            var s;
            var b = [];
            
            for ( var i = 0; i < a1.length; ++i ) {
                b[ i ] = [];
                for ( var j = 0; j < a2[ 0 ].length; ++j ) {
                    s = 0;
                    for ( var k = 0; k < a1[ 0 ].length; ++k ) {
                        s += a1[ i ][ k ] * a2[ k ][ j ];
                    }
                    b[ i ][ j ] = s;
                }
            }
            return b;
            */
            
            for ( var i = 0; i < n; ++i ) {
                a1[ i ] = [ // unwinded loop
                    a1[ i ][ 0 ] * a2[ 0 ][ 0 ]
                  + a1[ i ][ 1 ] * a2[ 1 ][ 0 ]
                  + a1[ i ][ 2 ] * a2[ 2 ][ 0 ]
                  + a1[ i ][ 3 ] * a2[ 3 ][ 0 ],
                    a1[ i ][ 0 ] * a2[ 0 ][ 1 ]
                  + a1[ i ][ 1 ] * a2[ 1 ][ 1 ]
                  + a1[ i ][ 2 ] * a2[ 2 ][ 1 ]
                  + a1[ i ][ 3 ] * a2[ 3 ][ 1 ],
                    a1[ i ][ 0 ] * a2[ 0 ][ 2 ]
                  + a1[ i ][ 1 ] * a2[ 1 ][ 2 ]
                  + a1[ i ][ 2 ] * a2[ 2 ][ 2 ]
                  + a1[ i ][ 3 ] * a2[ 3 ][ 2 ],
                    a1[ i ][ 0 ] * a2[ 0 ][ 3 ]
                  + a1[ i ][ 1 ] * a2[ 1 ][ 3 ]
                  + a1[ i ][ 2 ] * a2[ 2 ][ 3 ]
                  + a1[ i ][ 3 ] * a2[ 3 ][ 3 ]
                ];
            }
        }
    },
    triangle2d: function( a, b, c, clr, surface ) {
        var a = Alive.Transform.world2canvas( a, surface );
        var b = Alive.Transform.world2canvas( b, surface );
        var c = Alive.Transform.world2canvas( c, surface );
        
		var k = ( a[ 0 ] + b[ 0 ] + c[ 0 ] ) / 3; //x-pos of triangle center
		var l = ( a[ 1 ] + b[ 1 ] + c[ 1 ] ) / 3; //y-pos of triangle center
		
		//Find the nomalized vectors between the center and each of the corners
		var ak = Alive.Math.Vector.normalize( [ a[ 0 ] - k, a[ 1 ] - l, 0 ] );
		var bk = Alive.Math.Vector.normalize( [ b[ 0 ] - k, b[ 1 ] - l, 0 ] );
		var ck = Alive.Math.Vector.normalize( [ c[ 0 ] - k, c[ 1 ] - l, 0 ] );
		
		//Add the coresponding coordinates. Antialiasing will take care of subpixel coordinates
        a[ 0 ] += ak[ 0 ];
		b[ 0 ] += bk[ 0 ];
		c[ 0 ] += ck[ 0 ];
		
		a[ 1 ] += ak[ 1 ];
		b[ 1 ] += bk[ 1 ];
		c[ 1 ] += ck[ 1 ];
		
        surface.canvasSurface.beginPath();
        surface.canvasSurface.moveTo( a[ 0 ], a[ 1 ] );  
        surface.canvasSurface.lineTo( b[ 0 ], b[ 1 ] );  
        surface.canvasSurface.lineTo( c[ 0 ], c[ 1 ] ); 
        surface.canvasSurface.fillStyle = clr;
        surface.canvasSurface.fill();
        //surface.canvasSurface.strokeStyle = clr;
        //surface.canvasSurface.stroke();
    },
    Transform: {
        triangleMorphMatrix: function ( from, to ) {
            var A = from[ 0 ], B = from[ 1 ], C = from[ 2 ];
            var Ap = to[ 0 ], Bp = to[ 1 ], Cp = to[ 2 ];
            var Ax = A[ 0 ], Ay = A[ 1 ];
            var Bx = B[ 0 ], By = B[ 1 ];
            var Cx = C[ 0 ], Cy = C[ 1 ];
            var Axp = Ap[ 0 ], Ayp = Ap[ 1 ];
            var Bxp = Bp[ 0 ], Byp = Bp[ 1 ];
            var Cxp = Cp[ 0 ], Cyp = Cp[ 1 ];
            
            // The solution to a linear system of 6 equations in 6 unknowns. Mathematica form:
            // m = {{Ax, Ay, 0, 0, 1, 0}, {0, 0, Ax, Ay, 0, 1}, {Bx, By, 0, 0, 1, 0},
            //      {0, 0, Bx, By, 0, 1}, {Cx, Cy, 0, 0, 1, 0}, {0, 0, Cx, Cy, 0, 1}}
            // m.{p, q, r, s, g, d} == {Axp, Ayp, Bxp, Byp, Cxp, Cyp}
            // Solve[ %, {p, q, r, s, g, d}]
            //
            // This allows us to find the affine transformation required
            // It is critical that the points of the triangle are not colinear
            // Or else this system won't have a solution (yielding to division by zero and error)
            
            var p = -(-Axp*By+Axp*Cy+Ay*Bxp-Ay*Cxp-Bxp*Cy+By*Cxp)/(Ax*By-Ax*Cy-Ay*Bx+Ay*Cx+Bx*Cy-By*Cx);
            var q = -((Ax*Bxp-Ax*Cxp-Axp*Bx+Axp*Cx+Bx*Cxp-Bxp*Cx)/(-Ax*By+Ax*Cy+Ay*Bx-Ay*Cx-Bx*Cy+By*Cx));
            var r = -((-Ay*Byp+Ay*Cyp+Ayp*By-Ayp*Cy-By*Cyp+Byp*Cy)/(-Ax*By+Ax*Cy+Ay*Bx-Ay*Cx-Bx*Cy+By*Cx));
            var s = -((Ax*Byp-Ax*Cyp-Ayp*Bx+Ayp*Cx+Bx*Cyp-Byp*Cx)/(-Ax*By+Ax*Cy+Ay*Bx-Ay*Cx-Bx*Cy+By*Cx));
            var g = -((Ax*Bxp*Cy-Ax*By*Cxp-Axp*Bx*Cy+Axp*By*Cx+Ay*Bx*Cxp-Ay*Bxp*Cx)/(Ax*By-Ax*Cy-Ay*Bx+Ay*Cx+Bx*Cy-By*Cx));
            var d = -((Ax*By*Cyp-Ax*Byp*Cy-Ay*Bx*Cyp+Ay*Byp*Cx+Ayp*Bx*Cy-Ayp*By*Cx)/(-Ax*By+Ax*Cy+Ay*Bx-Ay*Cx-Bx*Cy+By*Cx));
            
            return [
                [ // linear transformation
                  [ p, q ],
                  [ r, s ]
                ],
                [ g, d ] // translation
            ];
        },
        world2canvas: function( v, surface ) {
            return [ ( v[ 0 ] + 1 ) * ( surface.width / 2 ), -( v[ 1 ] - 1 ) * ( surface.height / 2 ), v[ 2 ], v[ 3 ] ];
        },
        canvas2world: function( v, surface ) {
            var x = v[ 0 ];
            var y = v[ 1 ];
            
            return [ ( x / ( 2 * surface.width ) ) + 1, -( ( y / ( 2 * surface.height ) ) + 1 ) ];
        },
        project: function ( objData ) {
            // This function turns 3d points lying in our virtual space into
            // 2d points that we will draw on the screen
            // http://en.wikipedia.org/wiki/Perspective_projection#Perspective_projection
            var points = objData[ 0 ];
            var m = 1; // Math.PI / 2;
            var n = m * 1;
            
            var perspectiveProjection = 
                [  // x, y, z, w
                    [ m, 0, 0, 0 ], // m
                    [ 0, n, 0, 0 ], // n
                    [ 0, 0, ( Alive.CLIP_B + Alive.CLIP_F ) / ( Alive.CLIP_B - Alive.CLIP_F ), -( 2 * Alive.CLIP_B * Alive.CLIP_F ) / ( Alive.CLIP_B - Alive.CLIP_F ) ],
                    [ 0, 0, 0, 0 ]
                ];
            var vector2d = Alive.Math.Vector.copy( points );
            Alive.Math.mul( vector2d,
                 perspectiveProjection,
                 objData[ 2 ] // do not project the normal vectors
            );
            
            for ( var i = 0; i < vector2d.length; ++i ) {
                vector2d[ i ][ 0 ] = vector2d[ i ][ 0 ] / vector2d[ i ][ 3 ];
                vector2d[ i ][ 1 ] = vector2d[ i ][ 1 ] / vector2d[ i ][ 3 ];
            }
            
            return vector2d;
        },
        scale: function( objData, magnifier ) {
            // do not scale the normal vectors; after all they're unit vectors
            // TODO: normal vectors HAVE to be scaled if we're not scaling uniformly
            // also this is may cause problems to linear transformation compositions
            // e.g. if you try to compose rotation and scaling;
            // you can't know if you want the composition applied to normals or not!
            Alive.Math.mul( objData[ 0 ], Alive.Transform.scaleMatrix( magnifier ), objData[ 2 ] );
        },
        scaleMatrix: function( magnifier ) {
            // http://en.wikipedia.org/wiki/Scaling_(geometry)#Using_homogeneous_coordinates
            return [
                [ magnifier[ 0 ], 0, 0, 0 ],
                [ 0, magnifier[ 1 ], 0, 0 ],
                [ 0, 0, magnifier[ 2 ], 0 ],
                [ 0, 0, 0, 1 ]
            ];
        },
        translate: function( objData, offset ) {
            // move an object; we take advantage of the homogenous coordinates system of
            // the points to translate using a simple linear (affine) transformation.
            // Since the normal vectors do not need to be moved
            // we do not have to translate them.
            // But it's OK to translate them because a translation would not affect them
            // as they have a w coord of 0.
            Alive.Math.mul( objData[ 0 ], Alive.Transform.translationMatrix( offset ), objData[ 2 ] );
        },
        translationMatrix: function ( offset ) {
            // http://en.wikipedia.org/wiki/Translation_(geometry)
            return [
                // x, y, z, w
                [ 1, 0, 0, 0 ],
                [ 0, 1, 0, 0 ],
                [ 0, 0, 1, 0 ],
                [ offset[ 0 ], offset[ 1 ], offset[ 2 ], 1 ] // here's taking advantage of the w coord
            ];
        },
        rotate: function ( objData, angle ) {
            Alive.Math.mul( objData[ 0 ], Alive.Transform.rotationMatrix( angle ), objData[ 0 ].length ); // rotate the normal vectors as well
        },
        rotationMatrix: function( angle ) {
            var A = Math.cos( angle[ 0 ] ); // pitch
            var B = Math.sin( angle[ 0 ] );

            var C = Math.cos( angle[ 1 ] ); // yaw
            var D = Math.sin( angle[ 1 ] );

            var E = Math.cos( angle[ 2 ] ); // roll
            var F = Math.sin( angle[ 2 ] );
            
            // This is really the rotation matrix
            // produced by multiplying the three rotation matrices (pitch, yaw, and roll)
            // in order. This order matters! You may want to call
            // rotate() with all angles but one set to 0 and then call it
            // again to achieve a desired rotation.
            // http://en.wikipedia.org/wiki/Rotation_matrix#Three_dimensions
            return [
                [  C * E, -B * D * E + A * F        ,  A * D * E + B * F, 0 ],
                [ -C * F,  B * D * F + A * E        , -A * D * F + B * E, 0 ],
                [ -D    , -B * C                    ,  A * C            , 0 ], 
                [ 0     ,  0                        ,  0                , 1 ]
            ];
        },
        unit: function() {
            return [ [ 1, 0, 0, 0 ], [ 0, 1, 0, 0 ], [ 0, 0, 1, 0 ], [ 0, 0, 0, 1 ] ];
        }
    },
    Material: {
        create: function () {
            return {
                texture: 0,
                diffuse: [ 100, 150, 200 ],
                assignTexture: function ( texture ) {
                    this.texture = texture;
                }
            };
        }
    },
    Texture: {
        create: function ( url, primarycanvas, secondarycanvas, ready ) {
            var img = new Image();
            var primary = primarycanvas.getContext( '2d' );
            var secondary = secondarycanvas.getContext( '2d' );
            var loaded = false;
            
            img.onload = function () {
                loaded = true;
                primary.drawImage( img, 0, 0, 256, 256 );
                if ( typeof ready != 'undefined' ) {
                    ready();
                }
            };
            img.src = url;
            return {
                drawFace: function ( source, destination, surface ) {
                    if ( !loaded ) {
                        alert( 'Tried to use unloaded texture' );
                    }
                    // TODO: Optimize; somehow separate clipping and rotation
                    // and precompute clipping
                    var affine = Alive.Transform.triangleMorphMatrix( source, destination );
                    secondary.clearRect( 0, 0, 600, 600 );
                    secondary.save();
                    secondary.transform( affine[ 0 ][ 0 ][ 0 ], affine[ 0 ][ 1 ][ 0 ],
                                         affine[ 0 ][ 0 ][ 1 ], affine[ 0 ][ 1 ][ 1 ],
                                         affine[ 1 ][ 0 ], affine[ 1 ][ 1 ] );
                    secondary.beginPath();
                    secondary.moveTo( source[ 0 ][ 0 ], source[ 0 ][ 1 ] );
                    secondary.lineTo( source[ 1 ][ 0 ], source[ 1 ][ 1 ] );
                    secondary.lineTo( source[ 2 ][ 0 ], source[ 2 ][ 1 ] );
                    secondary.lineTo( source[ 0 ][ 0 ], source[ 0 ][ 1 ] );
                    secondary.clip();
                    secondary.drawImage( img, 0, 0, 256, 256, 0, 0, 600, 600 );
                    surface.canvasSurface.drawImage( secondarycanvas, 0, 0, 600, 600, 0, 0, 600, 600 );
                    secondary.restore();
                }
            };
        },
    },

    /*
    function centre( a, b, c ) {
        return [
                    ( a[ 0 ] + b[ 0 ] + c[ 0 ] ) / 3,
                    ( a[ 1 ] + b[ 1 ] + c[ 1 ] ) / 3,
                    ( a[ 2 ] + b[ 2 ] + c[ 2 ] ) / 3
               ];
    }
    */

    zorder: function ( universe ) {
        universe.faceref.sort( function ( a, b ) {
            // Painter's algorithm (priority fill):
            // Sort polygons by their barycenter and draw them back to front
            // http://en.wikipedia.org/wiki/Painter's_algorithm
            var points1 = universe.objects[ a[ 0 ] ].data[ 0 ];
            var points2 = universe.objects[ b[ 0 ] ].data[ 0 ];
            
            a = universe.objects[ a[ 0 ] ].data[ 1 ][ a[ 1 ] ];
            b = universe.objects[ b[ 0 ] ].data[ 1 ][ b[ 1 ] ];
            
            return ( points2[ b[ 0 ] ][ 2 ] + points2[ b[ 1 ] ][ 2 ] + points2[ b[ 2 ] ][ 2 ] ) / 3
                 - ( points1[ a[ 0 ] ][ 2 ] + points1[ a[ 1 ] ][ 2 ] + points1[ a[ 2 ] ][ 2 ] ) / 3;
        } );
    },
    drawUniverse: function( universe, surface ) {
        var triangle, point, a, b, c;
        var n, k;
        var projected = [];
        var obj, objidx;
        
        for ( var i = 0; i < universe.objects.length; ++i ) {
            projected[ i ] = Alive.Transform.project( universe.objects[ i ].data );
        }
        
        Alive.zorder( universe );
        
        for ( var i = 0; i < universe.faceref.length; ++i ) {
            objidx = universe.faceref[ i ][ 0 ];
            obj = universe.objects[ objidx ];
            triangle = obj.data[ 1 ][ universe.faceref[ i ][ 1 ] ];
            
            a = projected[ objidx ][ triangle[ 0 ] ];
            b = projected[ objidx ][ triangle[ 1 ] ];
            c = projected[ objidx ][ triangle[ 2 ] ];
            
            // Clipping part; determine which polygons don't need to be drawn
            // to speed up rendering
            // See http://en.wikipedia.org/wiki/Clipping_(computer_graphics)
            if ( a[ 2 ] < Alive.CLIP_F || b[ 2 ] < Alive.CLIP_F || c[ 2 ] < Alive.CLIP_F ) {
                continue; // front clipping
            }
            if ( a[ 2 ] > Alive.CLIP_B || b[ 2 ] > Alive.CLIP_B || c[ 2 ] > Alive.CLIP_B ) {
                continue; // back clipping
            }
            
            if ( Alive.CULLING ) {
                // frustum culling
                if ( 
                       ( a[ 0 ] < -1 && b[ 0 ] < -1 && c[ 0 ] < -1 ) // left
                    || ( a[ 0 ] > 1 && b[ 0 ] > 1 && c[ 0 ] > 1 ) // right
                    || ( a[ 1 ] < -1 && b[ 1 ] < -1 && c[ 1 ] < -1 ) // bottom
                    || ( a[ 1 ] > 1 && b[ 1 ] > 1 && c[ 1 ] > 1 ) // top
                    ) {
                    // TODO: Maybe we can do smarter culling of triangles.
                    // If a triangle has all its point outside the viewport but
                    // in different directions, it is possible that it may lie
                    // outside the viewport, so we should check for that
                    // Also, if a triangle partially lies within the viewport
                    // we may be able to cut it into pieces and only render the visible
                    // part. This may actually cause a speedup when textures are used.
                    continue;
                }
                
                // back-face culling
                // Do not draw the triangles that are facing away from the viewer
                // Which one is facing away and which not is determined by whether
                // the projected points are positioned clockwise or counter-clockwise.
                // We can figure out what's the case using the z-coordinate of the
                // cross product between vectors AB and AC.
                // 
                // http://en.wikipedia.org/wiki/Back-face_culling
                if ( ( b[ 0 ] - a[ 0 ] ) * ( c[ 1 ] - a[ 1 ] ) - ( b[ 1 ] - a[ 1 ] ) * ( c[ 0 ] - a[ 0 ] ) >= 0 ) {
                    continue;
                }
            }
            
            n = obj.data[ 0 ][ triangle[ 3 ] ];
            
            k = [
                    ( a[ 0 ] + b[ 0 ] + c[ 0 ] ) / 3,
                    ( a[ 1 ] + b[ 1 ] + c[ 1 ] ) / 3,
                    ( a[ 2 ] + b[ 2 ] + c[ 2 ] ) / 3
                ]; // center
            nk = Alive.Math.Vector.normalize( k );
            // calculate lighting
            angle = Math.abs( Alive.Math.Vector.dotproduct( n, nk ) );
            
            var red = Math.min( Math.ceil( Alive.LIGHT_AMBIENT[ 0 ] + angle * obj.material.diffuse[ 0 ] ), 255 );
            var green = Math.min( Math.ceil( Alive.LIGHT_AMBIENT[ 1 ] + angle * obj.material.diffuse[ 1 ] ), 255 );
            var blue = Math.min( Math.ceil( Alive.LIGHT_AMBIENT[ 2 ] + angle * obj.material.diffuse[ 2 ] ), 255 );
            
            as = Alive.Transform.world2canvas( a, surface );
            bs = Alive.Transform.world2canvas( b, surface );
            cs = Alive.Transform.world2canvas( c, surface );
            
            // draw coloured triangle
            Alive.triangle2d( a, b, c, '#' + red.toString( 16 ) + green.toString( 16 ) + blue.toString( 16 ), surface );
            
            // draw textured triangle
            if ( obj.material.texture != 0 ) {
                obj.material.texture.drawFace(
                    [ [ 0, 0 ], [ 255, 0 ], [ 255, 255 ] ],
                    [ as, bs, cs ], surface
                );
            }
        }
        
        ++Alive.frames;
    }
};
